"""Power Supply BSMP Entities.

Module for definitions of BSMP entities of power supply devices.

Documentation:

https://wiki-sirius.lnls.br/mediawiki/index.php/Machine:Power_Supplies
"""

from ...bsmp import Entity as _Entity
from ...bsmp import Entities as _Entities
from ...bsmp import Types as _Types

from .constants import ConstPSBSMP as _ConstPSBSMP


class Parameters(_Entity):
    """Power supply parameters."""

    _parameters = {
        # ----- class PS -----
        0: {'count': 64, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': False, 'Op': True},
        1: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': True, 'Op': False},
        2: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': True, 'Op': False},
        # ----- class Communication -----
        3: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': False, 'Op': True},
        4: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'bps',
            'init': True, 'Op': False},
        5: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': False, 'Op': True},
        6: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': True, 'Op': False},
        7: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': True, 'Op': False},
        8: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': True, 'Op': False},
        9: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
            'init': True, 'Op': False},
        10: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '%',
             'init': True, 'Op': False},
        # ----- class Control -----
        11: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'Hz',
             'init': True, 'Op': False},
        12: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': 'Hz',
             'init': True, 'Op': False},
        13: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        14: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': 'A/V',
             'init': False, 'Op': True},
        15: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': 'A/V',
             'init': False, 'Op': True},
        16: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '%',
             'init': False, 'Op': True},
        17: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '%',
             'init': False, 'Op': True},
        # ----- class PWM -----
        18: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'Hz',
             'init': True, 'Op': False},
        19: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'ns',
             'init': True, 'Op': False},
        20: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'pu',
             'init': False, 'Op': True},
        21: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'pu',
             'init': False, 'Op': True},
        22: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'pu',
             'init': False, 'Op': True},
        23: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'pu',
             'init': False, 'Op': True},
        24: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'pu',
             'init': False, 'Op': True},
        # ----- class HRADC -----
        25: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        26: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'MHz',
             'init': True, 'Op': False},
        27: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'Hz',
             'init': True, 'Op': False},
        28: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        29: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        30: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        31: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        32: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        # ----- class SigGen -----
        33: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        34: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        35: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'Hz',
             'init': False, 'Op': True},
        36: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'A/V/%',
             'init': False, 'Op': True},
        37: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'A/V/%',
             'init': False, 'Op': True},
        38: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        # ----- class WfmRef -----
        39: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        40: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
        41: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': 'Hz',
             'init': True, 'Op': False},
        42: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': 'A/V/%',
             'init': False, 'Op': True},
        43: {'count': 4, 'var_type': _Types.T_FLOAT, 'unit': 'A/V/%',
             'init': False, 'Op': True},
        # ----- class Analog Variables -----
        44: {'count': 64, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': False, 'Op': True},
        45: {'count': 64, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': False, 'Op': True},
        # ----- class Debounding manager -----
        46: {'count': 32, 'var_type': _Types.T_FLOAT, 'unit': 'us',
             'init': True, 'Op': False},
        47: {'count': 32, 'var_type': _Types.T_FLOAT, 'unit': 'us',
             'init': True, 'Op': False},
        48: {'count': 32, 'var_type': _Types.T_FLOAT, 'unit': 'us',
             'init': True, 'Op': False},
        49: {'count': 32, 'var_type': _Types.T_FLOAT, 'unit': 'us',
             'init': True, 'Op': False},
        # ---- Scope -----
        50: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': 'Hz',
             'init': True, 'Op': False},
        51: {'count': 1, 'var_type': _Types.T_FLOAT, 'unit': '',
             'init': True, 'Op': False},
    }

    def value_to_load(self, eid, value):
        """."""
        size = Parameters._parameters[eid]['count']
        var_types = [Parameters._parameters[eid]['var_type']] * size
        if eid == 0:
            # power supply name
            value = [float(ord(c)) for c in value]
        load = self._conv_value_to_load(var_types, size, value)
        return load

    def load_to_value(self, eid, load):
        """."""
        size = Parameters._parameters[eid]['count']
        var_types = [Parameters._parameters[eid]['var_type']] * size
        value = self._conv_load_to_value(var_types, load)
        if eid == 0:
            # power supply name
            value = [chr(int(v)) for v in value]
        return value

    @property
    def eids(self):
        """Return entities identifications."""
        return tuple(Parameters._parameters.keys())

    def __getitem__(self, key):
        """."""
        return Parameters._parameters[key]


class EntitiesPS(_Entities):
    """PS Entities."""

    _ps_variables = (
        # --- common variables
        {'eid': 0, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 1, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 2, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 3, 'waccess': False, 'count': 128, 'var_type': _Types.T_CHAR},
        {'eid': 4, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 5, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 6, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 7, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 8, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 9, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 10, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 11, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 12, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 13, 'waccess': False, 'count': 4, 'var_type': _Types.T_FLOAT},
        {'eid': 14, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 15, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 16, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 17, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 18, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 19, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 20, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 21, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 22, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 23, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 24, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 25, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 26, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 27, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 28, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        # --- undefined variables
        {'eid': 29, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        {'eid': 30, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
    )

    _ps_functions = (
        {'eid': _ConstPSBSMP.F_TURN_ON,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_TURN_OFF,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_OPEN_LOOP,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_CLOSE_LOOP,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SELECT_OP_MODE,
         'i_type': (_Types.T_ENUM,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_RESET_INTERLOCKS,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SET_COMMAND_INTERFACE,
         'i_type': (_Types.T_ENUM,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SET_SERIAL_TERMINATION,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_UNLOCK_UDC,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_LOCK_UDC,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_CFG_SOURCE_SCOPE,
         'i_type': (_Types.T_UINT32,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_CFG_FREQ_SCOPE,
         'i_type': (_Types.T_FLOAT,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_CFG_DURATION_SCOPE,
         'i_type': (_Types.T_FLOAT,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_ENABLE_SCOPE,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_DISABLE_SCOPE,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SYNC_PULSE,
         'i_type': (), 'o_type': ()},
        {'eid': _ConstPSBSMP.F_SET_SLOWREF,
         'i_type': (_Types.T_FLOAT,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SET_SLOWREF_FBP,
         'i_type': (_Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SET_SLOWREF_READBACK_MON,
         'i_type': (_Types.T_FLOAT,),
         'o_type': (_Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT,)},
        {'eid': _ConstPSBSMP.F_SET_SLOWREF_FBP_READBACK_MON,
         'i_type': (_Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT,),
         'o_type': (_Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT,)},
        {'eid': _ConstPSBSMP.F_SET_SLOWREF_READBACK_REF,
         'i_type': (_Types.T_FLOAT,),
         'o_type': (_Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT,)},
        {'eid': _ConstPSBSMP.F_SET_SLOWREF_FBP_READBACK_REF,
         'i_type': (_Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT,),
         'o_type': (_Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT,)},
        {'eid': _ConstPSBSMP.F_RESET_COUNTERS,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_CFG_WFMREF,
         'i_type': (_Types.T_UINT16, _Types.T_UINT16,
                    _Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SELECT_WFMREF,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_GET_WFMREF_SIZE,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT16, )},
        {'eid': _ConstPSBSMP.F_RESET_WFMREF,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_CFG_SIGGEN,
         'i_type': (_Types.T_ENUM, _Types.T_UINT16,
                    _Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SET_SIGGEN,
         'i_type': (_Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_ENABLE_SIGGEN,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_DISABLE_SIGGEN,
         'i_type': (), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SET_PARAM,
         'i_type': (_Types.T_PARAM, _Types.T_UINT16, _Types.T_FLOAT,),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_GET_PARAM,
         'i_type': (_Types.T_PARAM, _Types.T_UINT16,),
         'o_type': (_Types.T_FLOAT,)},
        {'eid': _ConstPSBSMP.F_SAVE_PARAM_EEPROM,
         'i_type': (_Types.T_PARAM, _Types.T_UINT16, _Types.T_UINT16,),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_LOAD_PARAM_EEPROM,
         'i_type': (_Types.T_PARAM, _Types.T_UINT16, _Types.T_UINT16,),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SAVE_PARAM_BANK,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_LOAD_PARAM_BANK,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SET_DSP_COEFFS,
         'i_type': (_Types.T_DSP_CLASS, _Types.T_UINT16,
                    _Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT,
                    _Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_GET_DSP_COEFF,
         'i_type': (_Types.T_DSP_CLASS, _Types.T_UINT16, _Types.T_UINT16,),
         'o_type': (_Types.T_FLOAT,)},
        {'eid': _ConstPSBSMP.F_SAVE_DSP_COEFFS_EEPROM,
         'i_type': (_Types.T_DSP_CLASS, _Types.T_UINT16, _Types.T_UINT16,),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_LOAD_DSP_COEFFS_EEPROM,
         'i_type': (_Types.T_DSP_CLASS, _Types.T_UINT16, _Types.T_UINT16,),
         'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_SAVE_DSP_MODULES_EEPROM,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_LOAD_DSP_MODULES_EEPROM,
         'i_type': (_Types.T_UINT16,), 'o_type': (_Types.T_UINT8,)},
        {'eid': _ConstPSBSMP.F_RESET_UDC,
         'i_type': (), 'o_type': ()},
    )

    _ps_curves = (
        {'eid': 0, 'waccess': True, 'count': 256,
         'nblocks': 16, 'var_type': _Types.T_FLOAT},
        {'eid': 1, 'waccess': True, 'count': 256,
         'nblocks': 16, 'var_type': _Types.T_FLOAT},
        {'eid': 2, 'waccess': False, 'count': 256,
         'nblocks': 16, 'var_type': _Types.T_FLOAT},
    )

    _ps_parameters = Parameters()

    def __init__(self):
        """Call super."""
        super().__init__(
            self._ps_variables, self._ps_curves, self._ps_functions)

    @property
    def parameters(self):
        """Return pwrsuppli parameters."""
        return EntitiesPS._ps_parameters


# --- Entities DCDC ---


class EntitiesFBP(EntitiesPS):
    """FBP-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FBP-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        # --- undefined Variables
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
        # --- mirror variables
        # ------ PS_STATUS
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT16},
        # ------ PS_SETPOINT
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        # ------ PS_REFERENCE
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        # ------ PS_SOFT_INTERLOCK
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 59, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 60, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 61, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        # ------ PS_HARD_INTERLOCK
        {'eid': 62, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 63, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 64, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 65, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        # ------ PS_I_LOAD
        {'eid': 66, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 67, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 68, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 69, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        # ------ V_PS_ALARMS
        {'eid': 70, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 71, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 72, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 73, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )

    _ps_curves = (
        {'eid': 0, 'waccess': True, 'count': 256,
         'nblocks': 4, 'var_type': _Types.T_FLOAT},
        {'eid': 1, 'waccess': True, 'count': 256,
         'nblocks': 4, 'var_type': _Types.T_FLOAT},
        {'eid': 2, 'waccess': False, 'count': 256,
         'nblocks': 4, 'var_type': _Types.T_FLOAT},
    )


class EntitiesFAC_DCDC(EntitiesPS):
    """FAC-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAC_DCDC-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )


class EntitiesFAC_2P4S_DCDC(EntitiesPS):
    """FAC_2P4S-type power supply entities (BO Dipoles)."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAC_2P4S-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 59, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 60, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 61, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 62, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 63, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 64, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 65, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 66, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 67, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 68, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 69, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 70, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 71, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 72, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 73, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 74, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 75, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 76, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 77, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 78, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 79, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 80, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 81, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 82, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )


class EntitiesFAC_2S_DCDC(EntitiesPS):
    """FAC_2S-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAC_2S_DCDC-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 59, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 60, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 61, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 62, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 63, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 64, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 65, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 66, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 67, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 68, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )


class EntitiesFAP(EntitiesPS):
    """FAP-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAP-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )


class EntitiesFAP_4P(EntitiesPS):
    """FAP_4P-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAP_4P-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 59, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 60, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 61, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 62, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 63, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 64, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 65, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 66, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 67, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 68, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 69, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 70, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 71, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 72, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 73, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 74, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 75, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 76, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 77, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 78, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 79, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 80, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 81, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 82, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 83, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 84, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 85, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 86, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 87, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 88, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 89, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 90, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 91, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 92, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 93, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 94, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 95, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 96, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 97, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 98, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 99, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 100, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 101, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 102, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 103, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 104, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 105, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 106, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 107, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 108, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 109, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 110, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 111, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 112, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 113, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 114, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 115, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 116, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 117, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 118, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 119, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 120, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 121, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 122, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )


class EntitiesFAP_2P2S(EntitiesPS):
    """FAP_2P2S-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAP_2P2S-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 59, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 60, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 61, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 62, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 63, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 64, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 65, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 66, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 67, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 68, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 69, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 70, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 71, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 72, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 73, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 74, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 75, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 76, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 77, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 78, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 79, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 80, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 81, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 82, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 83, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 84, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 85, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 86, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 87, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 88, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 89, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 90, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 91, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 92, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 93, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 94, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 95, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 96, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 97, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 98, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 99, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 100, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 101, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 102, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 103, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 104, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 105, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 106, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 107, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 108, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 109, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 110, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 111, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 112, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 113, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 114, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 115, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 116, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 117, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 118, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 119, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 120, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 121, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 122, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 123, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 124, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 125, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 126, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 127, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )


# --- Entities ACDC ---


class EntitiesFBP_DCLink(EntitiesPS):
    """FBP_DCLink-type power supplies entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FBP_DCLINK-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT8},
    )

    _ps_curves = ()


class EntitiesFAC_2S_ACDC(EntitiesPS):
    """FAC_2S_ACDC-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAC_2S_ACDC-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )

    _ps_curves = ()


class EntitiesFAC_2P4S_ACDC(EntitiesPS):
    """FAC_2P4S_ACDC-type power supply entities."""

    _ps_variables = EntitiesPS._ps_variables + (
        # --- FAC_2P4S_ACDC-specific variables
        {'eid': 31, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 32, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 33, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 34, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 35, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 36, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 37, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 38, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 39, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 40, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 41, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 42, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 43, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 44, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 45, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 46, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 47, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 48, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 49, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 50, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 51, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 52, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 53, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 54, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 55, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 56, 'waccess': False, 'count': 1, 'var_type': _Types.T_FLOAT},
        {'eid': 57, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
        {'eid': 58, 'waccess': False, 'count': 1, 'var_type': _Types.T_UINT32},
    )


    _ps_curves = ()



"""BSMP entities."""
import struct as _struct
import typing

import numpy as _np

from .types import BSMPType


class Entity:
    """BSMP entity."""

    def _conv_value(self, fmt: str, value) -> typing.List[str]:
        if fmt == '<c':
            return value
        else:
            return [chr(c) for c in _struct.pack(fmt, value)]

    def _check_type(self, var_type: BSMPType, value: typing.Union[str, int, float]):
        if not var_type.check(value):
            raise TypeError("{}, {}".format(var_type.type, value))

    def _calc_types_size(
        self,
        var_types: typing.Tuple[BSMPType],
    ) -> int:
        size = 0
        try:
            for var_type in var_types:
                size += var_type.size
        except TypeError:
            if isinstance(var_types, BSMPType):
                return var_types.size
            else:
                raise TypeError("Failed to calc types size, invalid type {}".format(var_types))

        return size

    def _conv_value_to_load(
        self,
        var_types: typing.Union[typing.Tuple[BSMPType], typing.List[BSMPType]],
        size: int,
        values: typing.Union[str, typing.List[str]]
    ) -> typing.List[str]:
        if len(var_types) > 1:
            load: typing.List[str] = []
            for idx, value in enumerate(values):
                self._check_type(var_types[idx], value)
                load += self._conv_value(var_types[idx].fmt, value)
            while len(load) < size:
                load += chr(0)
            return load
        else:
            self._check_type(var_types[0], values[0])
            return self._conv_value(var_types[0].fmt, values[0])

    def _conv_load_to_value(
        self,
        var_types: typing.Union[typing.Tuple[BSMPType], typing.List[BSMPType]],
        load: typing.List[str]
    ) -> typing.Union[str, float, int, typing.List[typing.Union[str, float, int]]]:
        """Return a value or a list of values unpacked according to the BMSPType.fmt"""
        # NOTE: optimize this critical function!
        _load = list(map(ord, load))
        if len(var_types) > 1:
            values = []
            offset = 0
            for var_type in var_types:
                datum = _load[offset:offset+var_type.size]
                values.append(_struct.unpack(var_type.fmt, bytes(datum))[0])
                offset += var_type.size
            return values
        return _struct.unpack(var_types[0].fmt, bytes(_load))[0]


class Variable(Entity):
    """BSMP variable."""

    def __init__(self, eid: int, waccess: bool, var_type: BSMPType, count: int = 1):
        """Set variable properties."""
        if (var_type.size * count) > 128 or (var_type.size * count) < 1:
            errstr = ('Variable size incorrect: eid:{}, '
                      'vtype_size:{}, count:{}').format(eid, var_type.size,
                                                        count)
            raise ValueError(errstr)
        super().__init__()  # NOTE: is it necessary?
        self.eid: int = eid
        self.waccess: bool = waccess
        self.size: int = (var_type.size * count)  # 1..128 bytes
        self.type: BSMPType = var_type

        self._var_types: typing.List[BSMPType] = [var_type for _ in range(count)]

    def load_to_value(self, load: typing.List[str]):
        """Parse value from load."""
        return self._conv_load_to_value(self._var_types, load)

    def value_to_load(self, value) -> typing.List[str]:
        """Convert value to load."""
        if not isinstance(value, (list, tuple, _np.ndarray)):
            value = [value, ]
        return self._conv_value_to_load(self._var_types, self.size, value)


class VariablesGroup(Entity):
    """BSMP variables group entity."""

    ALL: int = 0
    READ_ONLY: int = 1
    WRITEABLE: int = 2

    def __init__(
        self,
        eid: int,
        waccess: bool,
        variables: typing.List[Variable]
    ):
        """Set group parameter."""
        super().__init__()
        self.eid: int = eid
        self.waccess: bool = waccess
        self.size: int = len(variables)
        self.variables: typing.List[Variable] = variables

    def load_to_value(self, load: typing.List[str]) -> typing.List[typing.Union[str, float, int]]:
        """Parse value from load."""
        value: typing.List[typing.Union[str, float, int]] = []
        offset = 0
        for variable in self.variables:
            i, j = offset, offset + variable.size
            value.append(variable.load_to_value(load[i:j]))
            offset += variable.size
        return value

    def value_to_load(
        self,
        value: typing.List
    ) -> typing.List[str]:
        """Parse load from value."""
        if len(value) != self.size:
            # Value list must match the amount of variables
            return []
        load: typing.List[str] = []
        for i, variable in enumerate(self.variables):
            load.extend(variable.value_to_load(value[i]))
        return load

    def variables_size(self) -> int:
        """Return sum of variables size."""
        size = 0
        for variable in self.variables:
            size += variable.size
        return size


class Curve(Entity):
    """BSMP Curve entity."""

    def __init__(
        self,
        eid: int,
        waccess: bool,
        var_type: BSMPType,
        nblocks: int,
        count: int
    ):
        """Set curve properties."""
        super().__init__()
        self.eid: int = eid  # Entity ID
        self.waccess: bool = waccess
        self.size: int = (var_type.size * count)  # 1..128 bytes
        self.type: BSMPType = var_type
        self.nblocks: int = nblocks  # Number of blocks
        self.max_size_t_float: int = self.nblocks * (self.size // self.type.size)
        self._var_types: typing.List[BSMPType] = [var_type for _ in range(count)]

    def load_to_value(self, load: typing.List[str]):
        """Parse value from load."""
        _load = [ord(c) for c in load]
        values = []
        offset = 0
        for var_type in self._var_types:
            datum = _load[offset:offset+var_type.size]
            values.append(_struct.unpack(var_type.fmt, bytes(datum))[0])
            offset += var_type.size
            if offset >= len(_load):
                break
        return values

    def value_to_load(self, value) -> typing.List[str]:
        """Convert curve block number to load."""
        load: typing.List[str] = []
        for idx, val in enumerate(value):
            self._check_type(self._var_types[idx], val)
            load += self._conv_value(self._var_types[idx].fmt, val)
        return load

    def get_indices(self, data_length) -> typing.List[typing.Tuple[int, int]]:
        """Return list of indices corresponding to data blocks."""
        block_len = self.size // self.type.size
        nblocks = int(_np.ceil(data_length / block_len))
        indices: typing.List[typing.Tuple[int, int]] = []
        for i in range(nblocks-1):
            indices.append((block_len*i, block_len*(i+1)))
        indices.append((block_len*(nblocks-1), data_length))
        return indices


class Function(Entity):
    """BSMP function."""

    def __init__(
        self,
        eid: int,
        i_type: typing.Tuple[BSMPType],
        o_type: typing.Tuple[BSMPType]
    ):
        """Set function properties."""
        super().__init__()
        self.eid: int = eid
        i_size: int = self._calc_types_size(i_type)
        o_size: int = self._calc_types_size(o_type)
        if i_size < 0 or i_size > 64:
            raise ValueError("Input size {} is out of range".format(i_size))
        if o_size < 0 or o_size > 32:
            raise ValueError("Output size {} is out of range".format(o_size))

        self.i_size: int = i_size  # 0..64
        self.i_type: typing.Tuple[BSMPType] = i_type

        self.o_size: int = o_size  # 0..32
        self.o_type: typing.Tuple[BSMPType] = o_type

    def load_to_value(self, load: typing.Optional[typing.List[str]]):  # Parse output_size
        """Parse value from load."""
        if load is None or not load:
            return None
        return self._conv_load_to_value(self.o_type, load)

    def value_to_load(self, value) -> typing.List:
        """Convert value to load."""
        if value is None:
            return []
        if not isinstance(value, (list, tuple, _np.ndarray)):
            value = [value, ]
        return self._conv_value_to_load(self.i_type, self.i_size, value)


class Entities:
    """BSMP entities."""

    def __init__(
        self,
        variables: typing.Tuple[typing.Union[Variable, typing.Dict[str, typing.Any]]],
        curves: typing.Tuple[typing.Union[Curve, typing.Dict[str, typing.Any]]],
        functions: typing.Tuple[typing.Union[Function, typing.Dict[str, typing.Any]]]
    ):
        """Constructor."""
        # Get variables
        self._variables: typing.List[Variable] = []
        for variable in variables:

            if isinstance(variable, Variable):
                self.variables.append(variable)
                continue

            var_id = variable['eid']
            waccess = variable['waccess']
            var_type = variable['var_type']
            count = variable['count']
            self.variables.append(
                Variable(var_id, waccess, var_type, count)
            )

        # Standard groups
        self._groups: typing.List[VariablesGroup] = []
        self.reset_group()

        self._curves: typing.List[Curve] = []
        for curve in curves:

            if isinstance(curve, Curve):
                self.curves.append(curve)
                continue

            curve_id = curve['eid']
            waccess = curve['waccess']
            nblocks = curve['nblocks']
            count = curve['count']
            var_type = curve['var_type']
            self.curves.append(
                Curve(curve_id, waccess, var_type, nblocks, count))

        self._functions: typing.List[Function] = []
        for function in functions:

            if isinstance(function, Function):
                self.functions.append(function)
                continue

            func_id = function['eid']
            i_type = function['i_type']
            o_type = function['o_type']
            self.functions.append(Function(func_id, i_type, o_type))

    @property
    def variables(self) -> typing.List[Variable]:
        """Variables."""
        return self._variables

    @property
    def groups(self) -> typing.List[VariablesGroup]:
        """Groups."""
        return self._groups

    @property
    def curves(self) -> typing.List[Curve]:
        """Curves."""
        return self._curves

    @property
    def functions(self) -> typing.List[Function]:
        """Functions."""
        return self._functions

    def reset_group(self) -> None:
        """."""
        r_var = [var for var in self.variables if not var.waccess]
        w_var = [var for var in self.variables if var.waccess]
        self._groups = [
            VariablesGroup(0, False, self.variables),
            VariablesGroup(1, False, r_var),
            VariablesGroup(2, True, w_var),
        ]

    def add_group(self, var_ids: typing.List[int], waccess: bool = False):
        """Add group."""
        variables: typing.List[Variable] = []
        for var_id in var_ids:
            variables.append(self.variables[var_id])
        self.groups.append(VariablesGroup(len(self.groups), waccess, variables))

    def remove_all_groups_of_variables(self) -> None:
        """Remove all groups bigger than eid 2."""
        self._groups = self.groups[:3]

    def list_variables(self, group_id: int) -> typing.List[int]:
        """List variable ids."""
        return [var.eid for var in self.groups[group_id].variables]



"""Module to deal with correctors."""

import time as _time
import logging as _log
import traceback as _traceback

import numpy as _np
from PRUserial485 import EthBridgeClient

from .. import util as _util
from ..epics import PV as _PV
from ..thread import RepeaterThread as _Repeat
from ..pwrsupply.csdev import Const as _PSConst
from ..pwrsupply.bsmp.constants import ConstPSBSMP as _ConstPSBSMP
from ..timesys.csdev import Const as _TIConst
from ..search import HLTimeSearch as _HLTimesearch
from ..envars import VACA_PREFIX as LL_PREF
from ..namesys import SiriusPVName as _PVName
from ..devices import PSApplySOFB as _PSSOFBIOC

from ..pwrsupply.pssofb import PSSOFB as _PSSOFB


from .base_class import BaseClass as _BaseClass, \
    BaseTimingConfig as _BaseTimingConfig, compare_kicks as _compare_kicks

TIMEOUT = 0.05


class Corrector(_BaseTimingConfig):
    """Corrector class."""

    def __init__(self, corr_name):
        """Init method."""
        super().__init__(corr_name[:2])
        self._name = _PVName(corr_name)
        self._sp = None
        self._rb = None

    @property
    def name(self):
        """Corrector name."""
        return self._name

    @property
    def connected(self):
        """Status connected."""
        conn = super().connected
        pvs = (self._sp, self._rb)
        for pv in pvs:
            if not pv.connected:
                _log.debug('NOT CONN: ' + pv.pvname)
            conn &= pv.connected
        return conn

    @property
    def opmode_ok(self):
        """Opmode ok status."""
        return self.connected

    @property
    def opmode(self):
        """Opmode."""
        return 1

    @opmode.setter
    def opmode(self, val):
        """."""

    @property
    def sofbmode_ok(self):
        """Return SOFBMode Ok status."""
        return self.connected

    @property
    def sofbmode(self):
        """SOFBMode."""
        return 0

    @sofbmode.setter
    def sofbmode(self, val):
        """."""

    @property
    def state(self):
        """State."""
        pv = self._config_pvs_rb['PwrState']
        return pv.value == _PSConst.PwrStateSel.On if pv.connected else False

    @state.setter
    def state(self, boo):
        """."""
        val = _PSConst.PwrStateSel.On if boo else _PSConst.PwrStateSel.Off
        pv = self._config_pvs_sp['PwrState']
        self._config_ok_vals['PwrState'] = val
        if self.put_enable and pv.connected:
            pv.put(val, wait=False)

    @property
    def value(self):
        """Value."""
        if self._rb.connected:
            return self._rb.value

    @value.setter
    def value(self, val):
        """."""
        self._sp.put(val, wait=False)

    @property
    def refvalue(self):
        """."""
        return self.value


class RFCtrl(Corrector):
    """RF control class."""

    def __init__(self, acc):
        """Init method."""
        super().__init__(acc)
        self._name = self._csorb.RF_GEN_NAME
        opt = {'connection_timeout': TIMEOUT}
        pvpref = LL_PREF + ('-' if LL_PREF else '') + self._name
        self._sp = _PV(pvpref+':GeneralFreq-SP', **opt)
        self._rb = _PV(pvpref+':GeneralFreq-RB', **opt)
        self._config_ok_vals = {'PwrState': 1}
        self._config_pvs_sp = dict()
        #     'PwrState': _PV(pvpref+':PwrState-Sel', **opt)}
        self._config_pvs_rb = dict()
        #     'PwrState': _PV(pvpref+':PwrState-Sts', **opt)}

    @property
    def value(self):
        """Value."""
        if self._rb.connected:
            return self._rb.value

    @value.setter
    def value(self, freq):
        """."""
        delta_max = 200  # Hz
        freq0 = self.value
        if freq0 is None or freq is None:
            return
        delta = abs(freq-freq0)
        if delta < 0.1 or delta > 10000:
            return
        npoints = int(delta//delta_max) + 2
        freq_span = _np.linspace(freq0, freq, npoints)[1:]
        for i, freq in enumerate(freq_span, 1):
            self._sp.put(freq, wait=False)
            if i != freq_span.size:
                _time.sleep(1)

    @property
    def state(self):
        """State."""
        # TODO: database of RF GEN
        return True
        # pv = self._config_pvs_rb['PwrState']
        # return pv.value == 1 if pv.connected else False

    @state.setter
    def state(self, boo):
        """."""
        # TODO: database of RF GEN
        return
        # val = 1 if boo else 0
        # pv = self._config_pvs_sp['PwrState']
        # if pv.connected:
        #     self._config_ok_vals['PwrState'] = val
        #     pv.put(val, wait=False)


class CHCV(Corrector):
    """CHCV class."""

    def __init__(self, corr_name):
        """Init method."""
        super().__init__(corr_name)
        opt = {'connection_timeout': TIMEOUT}
        pvsp = self._name.substitute(
            prefix=LL_PREF, propty_name='Kick', propty_suffix='SP')
        pvrb = pvsp.substitute(propty_suffix='RB')
        pvref = pvsp.substitute(propty_name='KickRef', propty_suffix='Mon')
        self._sp = _PV(pvsp, **opt)
        self._rb = _PV(pvrb, **opt)
        self._ref = _PV(pvref, **opt)
        self._config_ok_vals = {
            'OpMode': _PSConst.OpMode.SlowRef,
            'PwrState': _PSConst.PwrStateSel.On,
            'SOFBMode': _PSConst.DsblEnbl.Dsbl}
        pvpref = self._name.substitute(prefix=LL_PREF)
        self._config_pvs_sp = {
            'OpMode': _PV(pvpref.substitute(propty='OpMode-Sel'), **opt),
            'PwrState': _PV(pvpref.substitute(propty='PwrState-Sel'), **opt),
            'SOFBMode': _PV(pvpref.substitute(propty='SOFBMode-Sel'), **opt)}
        self._config_pvs_rb = {
            'OpMode': _PV(pvpref.substitute(propty='OpMode-Sts'), **opt),
            'PwrState': _PV(pvpref.substitute(propty='PwrState-Sts'), **opt),
            'SOFBMode': _PV(pvpref.substitute(propty='SOFBMode-Sts'), **opt)}

    @property
    def opmode_ok(self):
        """Opmode ok status."""
        isok = self.opmode == self._config_ok_vals['OpMode']
        if not isok:
            msg = 'OpMode not Ok {0:s}'.format(self.name)
            _log.warning(msg)
        return isok

    @property
    def opmode(self):
        """Opmode."""
        pvobj = self._config_pvs_rb['OpMode']
        if not pvobj.connected:
            return None
        elif pvobj.value == _PSConst.States.SlowRefSync:
            return _PSConst.OpMode.SlowRefSync
        elif pvobj.value == _PSConst.States.SlowRef:
            return _PSConst.OpMode.SlowRef
        return pvobj.value

    @opmode.setter
    def opmode(self, val):
        """."""
        pvobj = self._config_pvs_sp['OpMode']
        self._config_ok_vals['OpMode'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def sofbmode_ok(self):
        """Opmode ok status."""
        isok = self.sofbmode == self._config_ok_vals['SOFBMode']
        if not isok:
            msg = 'SOFBMode not Ok {0:s}'.format(self.name)
            _log.warning(msg)
        return isok

    @property
    def sofbmode(self):
        """SOFBMode."""
        pvobj = self._config_pvs_rb['SOFBMode']
        if not pvobj.connected:
            return None
        return pvobj.value

    @sofbmode.setter
    def sofbmode(self, val):
        """."""
        pvobj = self._config_pvs_sp['SOFBMode']
        self._config_ok_vals['SOFBMode'] = val
        if self.put_enable and pvobj.connected:
            pvobj.put(val, wait=False)

    @property
    def connected(self):
        """Status connected."""
        conn = super().connected
        conn &= self._ref.connected
        if not self._ref.connected:
            _log.debug('NOT CONN: ' + self._ref.pvname)
        return conn

    @property
    def refvalue(self):
        """."""
        if self._ref.connected:
            return self._ref.value

    def configure(self):
        """Configure method."""
        if not self.connected:
            return False

        if not self.put_enable:
            return True
        val = self._config_ok_vals['SOFBMode']
        self.sofbmode = 0
        for k, pvo in self._config_pvs_sp.items():
            if k in self._config_ok_vals and k != 'SOFBMode':
                pvo.put(self._config_ok_vals[k], wait=False)
        self.sofbmode = val
        return True


class Septum(Corrector):
    """Septum class."""

    def __init__(self, corr_name):
        """Init method."""
        super().__init__(corr_name)
        opt = {'connection_timeout': TIMEOUT}
        pvsp = self._name.substitute(
            prefix=LL_PREF, propty_name='Kick', propty_suffix='SP')
        pvrb = pvsp.substitute(propty_suffix='RB')
        self._sp = _PV(pvsp, **opt)
        self._rb = _PV(pvrb, **opt)
        self._nominalkick = 98.55  # mrad
        self._config_ok_vals = {
            'Pulse': 1,
            'PwrState': _PSConst.PwrStateSel.On}
        pvpref = self._name.substitute(prefix=LL_PREF)
        self._config_pvs_sp = {
            'Pulse': _PV(pvpref.substitute(propty='Pulse-Sel'), **opt),
            'PwrState': _PV(pvpref.substitute(propty='PwrState-Sel'), **opt)}
        self._config_pvs_rb = {
            'Pulse': _PV(pvpref.substitute(propty='Pulse-Sts'), **opt),
            'PwrState': _PV(pvpref.substitute(propty='PwrState-Sts'), **opt)}

    @property
    def opmode_ok(self):
        """Opmode ok status."""
        return self.opmode == self._config_ok_vals['Pulse']

    @property
    def opmode(self):
        """Opmode."""
        pv = self._config_pvs_rb['Pulse']
        if not pv.connected:
            return None
        return pv.value

    @opmode.setter
    def opmode(self, val):
        """."""
        pv = self._config_pvs_sp['Pulse']
        self._config_ok_vals['Pulse'] = val
        if self.put_enable and pv.connected and pv.value != val:
            pv.put(val, wait=False)

    @property
    def value(self):
        """Value."""
        if self._rb.connected:
            val = self._rb.value
            if val is not None:
                return -(val + self._nominalkick) * 1e3

    @value.setter
    def value(self, val):
        """."""
        val = val/1e3 + self._nominalkick
        self._sp.put(-val, wait=False)


def get_corr(name):
    """."""
    if name.dis == 'PU':
        return Septum(name)
    else:
        return CHCV(name)


class TimingConfig(_BaseTimingConfig):
    """Timing configuration class."""

    def __init__(self, acc):
        """Init method."""
        super().__init__(acc)
        evt = self._csorb.evt_cor_name
        pref = LL_PREF + ('-' if LL_PREF else '')
        pref_name = pref + self._csorb.evg_name + ':' + evt
        trig = self._csorb.trigger_cor_name
        opt = {'connection_timeout': TIMEOUT}
        self._evt_sender = _PV(pref_name + 'ExtTrig-Cmd', **opt)
        src_val = self._csorb.CorrExtEvtSrc._fields.index(evt)
        src_val = self._csorb.CorrExtEvtSrc[src_val]
        self.EVT = src_val
        src_val = self._csorb.CorrExtEvtSrc._fields.index(
            self._csorb.clk_cor_name)
        src_val = self._csorb.CorrExtEvtSrc[src_val]
        self.CLK = src_val
        self._config_ok_vals = {
            'Mode': _TIConst.EvtModes.External,
            'Src': src_val, 'NrPulses': 1, 'Duration': 150.0, 'State': 1,
            'Polarity': 0,
            }
        if _HLTimesearch.has_delay_type(trig):
            self._config_ok_vals['RFDelayType'] = _TIConst.TrigDlyTyp.Manual
        pref_trig = pref + trig + ':'
        self._config_pvs_rb = {
            'Mode': _PV(pref_name + 'Mode-Sts', **opt),
            'Src': _PV(pref_trig + 'Src-Sts', **opt),
            'DelayRaw': _PV(pref_trig + 'DelayRaw-RB', **opt),
            'NrPulses': _PV(pref_trig + 'NrPulses-RB', **opt),
            'Duration': _PV(pref_trig + 'Duration-RB', **opt),
            'State': _PV(pref_trig + 'State-Sts', **opt),
            'Polarity': _PV(pref_trig + 'Polarity-Sts', **opt),
            }
        self._config_pvs_sp = {
            'Mode': _PV(pref_name + 'Mode-Sel', **opt),
            'Src': _PV(pref_trig + 'Src-Sel', **opt),
            'DelayRaw': _PV(pref_trig + 'DelayRaw-SP', **opt),
            'NrPulses': _PV(pref_trig + 'NrPulses-SP', **opt),
            'Duration': _PV(pref_trig + 'Duration-SP', **opt),
            'State': _PV(pref_trig + 'State-Sel', **opt),
            'Polarity': _PV(pref_trig + 'Polarity-Sel', **opt),
            }
        if _HLTimesearch.has_delay_type(trig):
            self._config_pvs_rb['RFDelayType'] = _PV(
                pref_trig + 'RFDelayType-Sts', **opt)
            self._config_pvs_sp['RFDelayType'] = _PV(
                pref_trig + 'RFDelayType-Sel', **opt)

    def send_evt(self):
        """Send event method."""
        self._evt_sender.value = 1

    @property
    def sync_type(self):
        """."""
        return self._config_pvs_rb['Src'].value

    @sync_type.setter
    def sync_type(self, value):
        """."""
        if value not in (self.EVT, self.CLK):
            return
        self._config_ok_vals['Src'] = value
        pvobj = self._config_pvs_sp['Src']
        if self.put_enable and pvobj.connected:
            pvobj.value = value

    @property
    def delayraw(self):
        """."""
        defv = 0
        pvobj = self._config_pvs_rb['DelayRaw']
        val = pvobj.value if pvobj.connected else defv
        return val if val else defv

    @delayraw.setter
    def delayraw(self, value):
        """."""
        pvobj = self._config_pvs_sp['DelayRaw']
        if self.put_enable and pvobj.connected:
            pvobj.value = int(value)

    @property
    def connected(self):
        """Status connected."""
        conn = super().connected
        conn &= self._evt_sender.connected
        if not self._evt_sender.connected:
            _log.debug('NOT CONN: ' + self._evt_sender.pvname)
        return conn


class BaseCorrectors(_BaseClass):
    """Base correctors class."""


class EpicsCorrectors(BaseCorrectors):
    """Class to deal with correctors."""

    TINY_INTERVAL = 0.005
    NUM_TIMEOUT = 1000
    PSSOFB_USE_IOC = False
    MAX_PROB = 5

    def __init__(self, acc, prefix='', callback=None, dipoleoff=False):
        """Initialize the instance."""
        super().__init__(acc, prefix=prefix, callback=callback)
        self._sofb = None
        self._sync_kicks = False
        self._acq_rate = 2

        self._names = self._csorb.ch_names + self._csorb.cv_names
        self._corrs = [get_corr(dev) for dev in self._names]
        if self.isring:
            self._corrs.append(RFCtrl(self.acc))

        self._use_pssofb = False
        self._wait_pssofb = False
        if self.acc == 'SI':
            self._ref_kicks = None
            self._ret_kicks = None
            self._func_ret = None
            self._prob = None
            if not EpicsCorrectors.PSSOFB_USE_IOC:
                self._pssofb = _PSSOFB(
                    EthBridgeClient, nr_procs=8, asynchronous=True,
                    sofb_update_iocs=True, dipoleoff=dipoleoff)
                self._pssofb.processes_start()
            else:
                self._pssofb = _PSSOFBIOC(
                    'SI', auto_mon=True, dipoleoff=dipoleoff)
            self.timing = TimingConfig(acc)
        self._corrs_thread = _Repeat(
            1/self._acq_rate, self._update_corrs_strength, niter=0)
        self._corrs_thread.start()

    @property
    def sofb(self):
        """."""
        return self._sofb

    @sofb.setter
    def sofb(self, sofb):
        self._sofb = sofb

    @property
    def corrs(self):
        """."""
        return self._corrs

    @property
    def use_pssofb(self):
        """."""
        return self._use_pssofb

    @use_pssofb.setter
    def use_pssofb(self, value):
        self.set_use_pssofb(value)

    def shutdown(self):
        """Shutdown Process."""
        if self.acc == 'SI':
            self._pssofb.processes_shutdown()

    def get_map2write(self):
        """Get the write methods of the class."""
        dbase = {
            'CorrConfig-Cmd': self.configure_correctors,
            'KickAcqRate-SP': self.set_kick_acq_rate,
            }
        if self.acc == 'SI':
            dbase['CorrPSSOFBEnbl-Sel'] = self.set_use_pssofb
            dbase['CorrPSSOFBWait-Sel'] = self.set_wait_pssofb
            dbase['CorrSync-Sel'] = self.set_corrs_mode
        return dbase

    def apply_kicks(self, values):
        """Apply kicks.

        Will return -2 if there is a problem with some PS.
        Will return -1 if PS did not finish applying last kick.
        Will return 0 if all previous kick were implemented.
        Will return >0 indicating how many previous kicks were not implemented.
        """
        if self.acc == 'BO':
            msg = 'ERR: Cannot correct Orbit in Booster. Use Ramp Interface!'
            self._update_log(msg)
            _log.error(msg[5:])
            return 0

        if self.acc == 'SI' and self._use_pssofb:
            return self.apply_kicks_pssofb(values)

        strn = '    TIMEIT: {0:20s} - {1:7.3f}'
        _log.info('    TIMEIT: BEGIN')
        time1 = _time.time()

        not_nan_idcs = ~_np.isnan(values)
        # Send correctors setpoint
        for i, corr in enumerate(self._corrs):
            if not_nan_idcs[i]:
                self.put_value_in_corr(corr, values[i])
        time2 = _time.time()
        _log.info(strn.format('send sp:', 1000*(time2-time1)))

        # Wait for readbacks to be updated
        if self._timed_out(values, mode='ready'):
            return -1
        time3 = _time.time()
        _log.info(strn.format('check ready:', 1000*(time3-time2)))

        # Send trigger signal for implementation
        self.send_evt()
        time4 = _time.time()
        _log.info(strn.format('send evt:', 1000*(time4-time3)))

        # Wait for references to be updated
        self._timed_out(values, mode='applied')
        time5 = _time.time()
        _log.info(strn.format('check applied:', 1000*(time5-time4)))
        _log.info('    TIMEIT: END')
        return 0

    def apply_kicks_pssofb(self, values):
        """Apply kicks with PSSOFB.

        Will return -2 if there is a problem with some PS.
        Will return -1 if PS did not finish applying last kick.
        Will return 0 if all previous kick were implemented.
        Will return >0 indicating how many previous kicks were not implemented.
        """
        if not self._pssofb.is_ready():
            msg = 'ERR: PSSOFB not ready!'
            self._update_log(msg)
            _log.error(msg[5:])
            return -1

        # Send correctors setpoint
        ret_kicks = self._pssofb.sofb_kick_readback_ref
        func_ret = self._pssofb.sofb_func_return
        self._ret_kicks = ret_kicks
        self._func_ret = func_ret

        self._set_ref_kicks(values)
        # NOTE: values will contain NaNs when corrector is not in the loop, as
        # a mean to not communicate with it. Pass them to PSSOFB so it handles
        # this appropriately (actually it is the PSBSMP class that will handle
        # NaNs):
        self._pssofb.bsmp_sofb_kick_set(values[:-1])
        if not _np.isnan(values[-1]):
            self.put_value_in_corr(self._corrs[-1], values[-1])

        if self._wait_pssofb and not self._pssofb.wait(timeout=1):
            msg = 'ERR: PSSOFB timed out: Worker is not Done!'
            self._update_log(msg)
            _log.error(msg[5:])

        # compare kicks to check if there is something wrong
        ret = self._compare_kicks_pssofb(ret_kicks, func_ret)

        # Send trigger signal for implementation
        self.send_evt()

        return ret

    def _set_ref_kicks(self, values):
        new_ref = _np.array(self._ref_kicks[-1])
        not_nan = ~_np.isnan(values)
        new_ref[not_nan] = values[not_nan]
        self._ref_kicks.append(new_ref)
        del self._ref_kicks[0]

    def put_value_in_corr(self, corr, value):
        """Put value in corrector method."""
        if not corr.connected:
            msg = 'ERR: ' + corr.name + ' not connected.'
            self._update_log(msg)
            _log.error(msg[5:])
        elif not corr.state:
            msg = 'ERR: ' + corr.name + ' is off.'
            self._update_log(msg)
            _log.error(msg[5:])
        elif not corr.opmode_ok:
            msg = 'ERR: ' + corr.name + ' mode not configured.'
            self._update_log(msg)
            _log.error(msg[5:])
        else:
            corr.value = value

    def send_evt(self):
        """Send event method."""
        if self.acc != 'SI' or self._sync_kicks != self._csorb.CorrSync.Event:
            return
        if not self.timing.connected:
            msg = 'ERR: timing disconnected.'
            self._update_log(msg)
            _log.error(msg[5:])
            return
        elif not self.timing.is_ok:
            msg = 'ERR: timing not configured.'
            self._update_log(msg)
            _log.error(msg[5:])
            return
        self.timing.send_evt()

    def get_strength(self):
        """Get the correctors strengths."""
        corr_values = _np.zeros(self._csorb.nr_corrs, dtype=float)
        if self.acc == 'SI' and self._use_pssofb:
            return self._ref_kicks[-1].copy()

        for i, corr in enumerate(self._corrs):
            if corr.connected and corr.value is not None:
                corr_values[i] = corr.refvalue
            else:
                msg = 'ERR: Failed to get value from '
                msg += corr.name
                self._update_log(msg)
                _log.error(msg[5:])
        return corr_values

    def set_kick_acq_rate(self, value):
        """Set kick caq rate method."""
        self._acq_rate = value
        self._corrs_thread.interval = 1/value
        self.run_callbacks('KickAcqRate-RB', value)
        return True

    def _update_corrs_strength(self):
        try:
            corr_vals = self.get_strength()
            self.run_callbacks('KickCH-Mon', corr_vals[:self._csorb.nr_ch])
            self.run_callbacks(
                'KickCV-Mon', corr_vals[self._csorb.nr_ch:self._csorb.nr_chcv])
            if self.isring and corr_vals[-1] > 0:
                # NOTE: I have to check whether the RF frequency is larger
                # than zero not to take the inverse of 0. It will be 0 in case
                # there is a failure to get the RF frequency from its PV.
                rfv = corr_vals[-1]
                circ = 1/rfv * self._csorb.harm_number * 299792458
                self.run_callbacks('KickRF-Mon', rfv)
                self.run_callbacks('OrbLength-Mon', circ)
        except Exception as err:
            self._update_log('ERR: ' + str(err))
            _log.error(_traceback.format_exc())

    def set_corrs_mode(self, value):
        """Set mode of CHs and CVs method. Only called when acc==SI."""
        if value not in self._csorb.CorrSync:
            return False
        self._sync_kicks = value
        val = _PSConst.OpMode.SlowRefSync
        if value == self._csorb.CorrSync.Off:
            self.set_timing_delay(0)
            val = _PSConst.OpMode.SlowRef
        elif value == self._csorb.CorrSync.Event:
            self.set_timing_delay(0)
            self.timing.sync_type = self.timing.EVT
        else:
            self.set_timing_delay(self._csorb.CORR_DEF_DELAY)
            self.timing.sync_type = self.timing.CLK

        mask = self._get_mask()
        for i, corr in enumerate(self._corrs[:-1]):
            if mask[i] and not corr.connected:
                msg = 'ERR: Failed to configure '
                msg += corr.name
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            corr.put_enable = mask[i]
            corr.opmode = val

        strsync = self._csorb.CorrSync._fields[self._sync_kicks]
        msg = 'Synchronization set to {0:s}'.format(strsync)
        self._update_log(msg)
        _log.info(msg)
        self.run_callbacks('CorrSync-Sts', value)
        return True

    def set_timing_delay(self, value):
        """."""
        frf = 499666 / 4  # [kHz]
        raw = int(value * frf)
        self.timing.delayraw = raw
        return True

    def set_use_pssofb(self, val):
        """."""
        if self.acc != 'SI':
            return False
        if val not in self._csorb.CorrPSSOFBEnbl:
            return False
        if val != self._use_pssofb and val == self._csorb.CorrPSSOFBEnbl.Enbld:
            kicks = self.get_strength()
            self._ref_kicks = [kicks.copy(), kicks.copy(), kicks.copy()]
            self._prob = _np.zeros(kicks.size, dtype=_np.int8)
            # initialize PSSOFB State
            self._pssofb.bsmp_sofb_kick_set(kicks[:-1])
        self._use_pssofb = val

        # NOTE: We need this time to avoid problems in the Correctors IOCs.
        # We noticed that without this sleep, in ramdom manner, the IOCs from
        # some correctors do not turn off the SOFB mode when requested, which
        # required the reboot of the IOC to get back control of the corrs.
        # We noticed this problem only happens when we use PSSOFB to control
        # the correctors. This means that PSConnSOFB, which doesn't use
        # multiprocessing, is problem free somehow.
        # We don't understand the reason for this sleep to solve the problem
        # and neither why the problem even occurs...
        _time.sleep(0.2)

        mask = self._get_mask()
        for i, corr in enumerate(self._corrs[:-1]):
            if mask[i] and not corr.connected:
                msg = 'ERR: Failed to configure '
                msg += corr.name
                self._update_log(msg)
                _log.error(msg[5:])
                continue
            corr.put_enable = mask[i]
            corr.sofbmode = val

        self.run_callbacks('CorrPSSOFBEnbl-Sts', val)
        return True

    def set_wait_pssofb(self, val):
        """."""
        if self.acc != 'SI':
            return False
        if val not in self._csorb.CorrPSSOFBWait:
            return False
        self._wait_pssofb = val
        self.run_callbacks('CorrPSSOFBWait-Sts', val)
        return True

    def configure_correctors(self, _):
        """Configure correctors method."""
        corrs = self._get_used_corrs(include_rf=True)
        for corr in corrs:
            if not corr.connected:
                msg = 'ERR: Failed to configure '
                msg += corr.name
                self._update_log(msg)
                _log.error(msg[5:])
                continue
            # Do not configure opmode in BO corrs because they are ramping.
            if self.acc == 'BO':
                corr.state = True
                continue
            corr.configure()
        if not self.isring:
            return True

        if self.acc == 'SI' and self._sync_kicks != self._csorb.CorrSync.Off:
            if not self.timing.configure():
                msg = 'ERR: Failed to configure timing'
                self._update_log(msg)
                _log.error(msg[5:])
        return True

    def _update_status(self):
        status = 0b0000111
        if self.acc == 'SI':
            status = 0b1111111
        elif self.isring:
            status = 0b0011111

        chcvs = self._get_used_corrs(include_rf=False)

        status = _util.update_bit(
            status, bit_pos=0,
            bit_val=not all(corr.connected for corr in chcvs))
        # Do not check mode of BO correctors because they are ramping.
        opmode_ok = True
        if self.acc != 'BO':
            opmode_ok = all(corr.opmode_ok for corr in chcvs)
        if self.acc == 'SI':
            opmode_ok &= all(corr.sofbmode_ok for corr in chcvs)
        status = _util.update_bit(status, bit_pos=1, bit_val=not opmode_ok)
        status = _util.update_bit(
            status, bit_pos=2,
            bit_val=not all(corr.state for corr in chcvs))
        if self.acc == 'SI' and self._sync_kicks != self._csorb.CorrSync.Off:
            tim_conn = self.timing.connected
            tim_conf = self.timing.is_ok
        else:
            tim_conn = tim_conf = True
        status = _util.update_bit(status, bit_pos=3, bit_val=not tim_conn)
        status = _util.update_bit(status, bit_pos=4, bit_val=not tim_conf)
        if self.isring:
            rfctrl = self._corrs[-1]
            status = _util.update_bit(
                status, bit_pos=5, bit_val=not rfctrl.connected)
            status = _util.update_bit(
                status, bit_pos=6, bit_val=not rfctrl.state)
        self._status = status
        self.run_callbacks('CorrStatus-Mon', status)

    def _timed_out(self, values, mode='ready'):
        okg = [False, ] * len(self._corrs)
        for _ in range(self.NUM_TIMEOUT):
            for i, corr in enumerate(self._corrs):
                if okg[i]:
                    continue
                if _np.isnan(values[i]):
                    okg[i] = True
                    continue
                val = corr.value if mode == 'ready' else corr.refvalue
                okg[i] = val is not None and _compare_kicks(values[i], val)
            if all(okg):
                return False
            _time.sleep(self.TINY_INTERVAL)

        self._print_guilty(okg, mode=mode)
        return True

    def _compare_kicks_pssofb(self, values, fret):
        mask = self._get_mask()
        rfc = self._corrs[-1]
        ref_vals = self._ref_kicks[0]

        # NOTE: In case there is some problem in the return value, apart from
        # DSP_TIMEOUT, which we consider not a fatal problem, return code
        # error -2 so that the loop can be opened:
        timeout = _np.zeros(ref_vals.size, dtype=bool)
        timeout[:-1] = fret == _ConstPSBSMP.ACK_DSP_TIMEOUT
        comm_prob = fret != _ConstPSBSMP.ACK_OK
        comm_prob &= ~timeout[:-1]
        comm_prob &= mask[:-1]  # If corrector is not in the loop, ignore it.
        if _np.any(comm_prob):
            self._print_guilty(~comm_prob, mode='prob_code', fret=fret)
            return -2

        curr_vals = _np.zeros(self._csorb.nr_corrs, dtype=float)
        curr_vals[:-1] = values
        curr_vals[-1] = rfc.value

        # NOTE: If the return value is zero, it might mean the corrector had a
        # problem. In this case we return -2 so the main correction loop can
        # exit and give back the control to the IOC.
        iszero = _np.equal(curr_vals, 0.0)
        iszero_ref = _compare_kicks(ref_vals, 0)
        prob = iszero & ~(iszero_ref)
        prob &= mask  # If corrector is not in the loop, ignore it.
        # For debugging:
        # if _np.any(prob):
        #     self._print_guilty(
        #         ~prob, mode='prob_curr', currs=curr_vals, refs=ref_vals)

        # Only acuse problem if it repeats for MAX_PROB consecutive times:
        # Because in some cases of previous unsuccessful applications, the
        # current value will be different from this reference.
        self._prob[prob] += 1
        self._prob[~prob] = 0
        if _np.any(self._prob >= self.MAX_PROB):
            self._print_guilty(
                ~prob, mode='prob_curr', currs=curr_vals, refs=ref_vals)
            return -2

        diff = _np.logical_not(_compare_kicks(curr_vals, ref_vals))

        # NOTE: I don't remember why it only matters when differences are due
        # to Timeout error:
        diff &= timeout

        # If corrector is not in the loop, ignore it:
        diff &= mask
        return _np.sum(diff)

    def _print_guilty(
            self, okg, mode='ready', fret=None, currs=None, refs=None):
        msg_tmpl = 'ERR: timeout {0:3s}: {1:s}'
        data = [tuple(), ] * len(self._corrs)
        if mode == 'prob_code':
            msg_tmpl = 'ERR: {0:s} --> {1:s}: code={2:d}'
            data = zip(fret)
        elif mode == 'prob_curr':
            msg_tmpl = 'ERR: {0:s} --> {1:s}: curr={2:.4g}, ref={3:.4g}'
            data = zip(currs, refs)
        elif mode == 'diff':
            msg_tmpl = 'ERR: Corrector {1:s} diff from setpoint!'
        for oki, corr, args in zip(okg, self._corrs, data):
            if not oki:
                msg = msg_tmpl.format(mode, corr.name, *args)
                self._update_log(msg)
                _log.error(msg[5:])

    def _get_mask(self):
        if self.sofb is not None and self.sofb.matrix is not None:
            mask = self.sofb.matrix.corrs_enbllist
        else:
            mask = _np.ones(len(self._corrs), dtype=bool)
        return mask

    def _get_used_corrs(self, include_rf=False):
        corrs = []
        nrc = None if include_rf else self._csorb.nr_chcv
        mask = self._get_mask()[:nrc]
        for i, corr in enumerate(self._corrs[:nrc]):
            corr.put_enable = mask[i]
            if mask[i]:
                corrs.append(corr)
        return corrs



"""RF Diag App."""

from ... import csdev as _csdev


class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    DIAG_STATUS_LABELS_BORF = (
        'Disconnected',
        'Sirius Interlock',
        'LLRF Interlock',
        'Ramp Disabled',
        'Ramp Not Ready',
    )

    DIAG_STATUS_LABELS_SIRF = (
        'Disconnected',
        'Sirius Interlock',
        'LLRF Interlock',
        'Amplitude Error Out Of Tolerance',
        'Phase Error Out Of Tolerance',
        'Detune Error Out of Tolerance',
    )


_et = ETypes  # syntactic sugar


class Const(_csdev.Const):
    """Local constant types."""

    BO_DEV = 'BO-05D:RF-P5Cav'
    SI_DEV = 'SI-02SB:RF-P7Cav'
    ALL_DEVICES = [BO_DEV, SI_DEV]

    SI_SL_ERRTOL_AMP = 1  # [mV]
    SI_SL_ERRTOL_PHS = 1e-1  # [DEG]
    SI_SL_ERRTOL_DTU = 3  # [DEG]


_c = Const  # syntactic sugar


def get_rf_diag_status_labels(device):
    """Return Diag Status Labels enum."""
    if 'BO' in device:
        return _et.DIAG_STATUS_LABELS_BORF
    elif 'SI' in device:
        return _et.DIAG_STATUS_LABELS_SIRF
    raise ValueError('Labels not defined to '+device+'.')


def get_rf_diag_propty_database(device):
    """Return property database of diagnostics for RF devices."""
    enums = get_rf_diag_status_labels(device)
    dbase = {
        'DiagVersion-Cte': {'type': 'str', 'value': 'UNDEF'},
        'DiagStatus-Mon': {'type': 'int', 'value': 0,
                           'hilim': 1, 'hihi': 1, 'high': 1,
                           'low': -1, 'lolo': -1, 'lolim': -1},
        'DiagStatusLabels-Cte': {'type': 'string', 'count': len(enums),
                                 'value': enums},
    }
    if device.startswith('SI'):
        dbase.update({
            'DiagAmpErrSts-Mon': {'type': 'int', 'value': 0,
                                  'hilim': 1, 'hihi': 1, 'high': 1,
                                  'low': -1, 'lolo': -1, 'lolim': -1},
            'DiagPhsErrSts-Mon': {'type': 'int', 'value': 0,
                                  'hilim': 1, 'hihi': 1, 'high': 1,
                                  'low': -1, 'lolo': -1, 'lolim': -1},
            'DiagDTuneErrSts-Mon': {'type': 'int', 'value': 0,
                                    'hilim': 1, 'hihi': 1, 'high': 1,
                                    'low': -1, 'lolo': -1, 'lolim': -1},
        })
    dbase = _csdev.add_pvslist_cte(dbase, 'Diag')
    return dbase



"""Define Insertion Devices."""

import time as _time
import numpy as _np

from ..namesys import SiriusPVName as _SiriusPVName
from ..search import IDSearch as _IDSearch
from ..magnet.idffwd import APUFFWDCalc as _APUFFWDCalc

from .device import Device as _Device
from .device import Devices as _Devices
from .device import DeviceApp as _DeviceApp
from .psconv import StrengthConv as _StrengthConv


class APU(_Device):
    """Insertion Device APU."""

    class DEVICES:
        """."""

        APU22_06SB = 'SI-06SB:ID-APU22'
        APU22_07SP = 'SI-07SP:ID-APU22'
        APU22_08SB = 'SI-08SB:ID-APU22'
        APU22_09SA = 'SI-09SA:ID-APU22'
        APU58_11SP = 'SI-11SP:ID-APU58'
        ALL = (APU22_06SB, APU22_07SP, APU22_08SB, APU22_09SA,
               APU58_11SP, )

    _properties = (
        'DevCtrl-Cmd', 'Moving-Mon',
        'PhaseSpeed-SP', 'PhaseSpeed-Mon',
        'Phase-SP', 'Phase-Mon',
        'Kx-SP', 'Kx-Mon',
    )

    _DEF_TIMEOUT = 10  # [s]
    _CMD_MOVE = 3
    _MOVECHECK_SLEEP = 0.1  # [s]

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in APU.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=APU._properties, auto_mon=True)

    @property
    def phase(self):
        """Return APU phase [mm]."""
        return self['Phase-Mon']

    @phase.setter
    def phase(self, value):
        """Set APU phase [mm]."""
        self['Phase-SP'] = value

    @property
    def phase_sp(self):
        """Return APU phase SP [mm]."""
        return self['Phase-SP']

    @property
    def phase_speed(self):
        """Return APU phase speed [mm/s]."""
        return self['PhaseSpeed-Mon']

    @phase_speed.setter
    def phase_speed(self, value):
        """Set APU phase_speed [mm/s]."""
        self['PhaseSpeed-SP'] = value

    @property
    def phase_speed_sp(self):
        """Return APU phase speed SP [mm/s]."""
        return self['PhaseSpeed-SP']

    @property
    def idkx(self):
        """Return APU Kx."""
        return self['Kx-SP']

    @idkx.setter
    def idkx(self, value):
        """Set APU Kx."""
        self['Kx-SP'] = value

    @property
    def is_moving(self):
        """Return True if phase is changing."""
        return round(self['Moving-Mon']) == 1

    def cmd_move(self, timeout=_DEF_TIMEOUT):
        """."""
        self['DevCtrl-Cmd'] = APU._CMD_MOVE
        return True

    def wait_move(self):
        """Wait for phase movement to complete."""
        _time.sleep(APU._MOVECHECK_SLEEP)
        while self.is_moving:
            _time.sleep(APU._MOVECHECK_SLEEP)


class IDCorrectors(_DeviceApp):
    """."""

    DEVICES = APU.DEVICES

    def __init__(self, devname):
        """."""
        devname = _SiriusPVName(devname)

        # check if device exists
        if devname not in IDCorrectors.DEVICES.ALL:
            raise NotImplementedError(devname)

        # get correctors names
        self._psnames_orb = _IDSearch.conv_idname_2_orbitcorr(devname)

        # get deviceapp properties
        properties, \
            self._orb_sp, self._orb_rb, self._orb_refmon, self._orb_mon = \
            self._get_properties()

        # call base class constructor
        super().__init__(properties=properties, devname=devname)

    @property
    def orbitcorr_psnames(self):
        """Return orbit corrector names."""
        return self._psnames_orb

    @property
    def orbitcorr_current(self):
        """Return orbit SOFBCurrent Mon."""
        return self[self._orb_mon]

    @orbitcorr_current.setter
    def orbitcorr_current(self, value):
        """Set orbit SOFBCurrent SP."""
        self[self._orb_sp] = value

    @property
    def orbitcorr_current_sp(self):
        """Return orbit SOFBCurrent setpoint."""
        return self[self._orb_sp]

    @property
    def orbitcorr_current_rb(self):
        """Return orbit SOFBCurrent readback."""
        return self[self._orb_rb]

    @property
    def orbitcorr_current_mon(self):
        """Return orbit SOFBCurrent monitor."""
        return self[self._orb_mon]

    def _get_properties(self):
        corrname = self._psnames_orb[0]
        orb_sp = corrname + ':SOFBCurrent-SP'
        orb_rb = corrname + ':SOFBCurrent-RB'
        orb_refmon = corrname + ':SOFBCurrentRef-Mon'
        orb_mon = corrname + ':SOFBCurrent-Mon'
        properties = (
            orb_sp, orb_rb, orb_refmon, orb_mon
        )
        return properties, orb_sp, orb_rb, orb_refmon, orb_mon


class APUFeedForward(_Devices):
    """Insertion Device APU FeedForward."""

    DEVICES = APU.DEVICES

    def __init__(self, devname):
        """."""
        # create APU device
        self._apu = APU(devname)

        # create IDCorrectors
        self._idcorrs = IDCorrectors(devname)

        # create FFWDCalc
        self._ffwdcalc = _APUFFWDCalc(devname)

        # create normalizers
        self._strenconv_chs, self._strenconv_cvs = self._create_strenconv()

        # call base class constructor
        devices = (
            self._apu, self._idcorrs,
            self._strenconv_chs, self._strenconv_cvs)
        super().__init__(devname, devices)

        # bumps
        self._posx, self._angx, self._posy, self._angy = \
            self._init_posang()

    @property
    def apu(self):
        """Return APU device."""
        return self._apu

    @property
    def correctors(self):
        """Return IDCorrectors device."""
        return self._idcorrs

    @property
    def ffwdcalc(self):
        """."""
        return self._ffwdcalc

    @property
    def posx(self):
        """Return posx bump value."""
        return self._posx

    @posx.setter
    def posx(self, value):
        """Return posx bump value."""
        self._posx = value

    @property
    def angx(self):
        """Return angx bump value."""
        return self._angx

    @angx.setter
    def angx(self, value):
        """Return angx bump value."""
        self._angx = value

    @property
    def posy(self):
        """Return posy bump value."""
        return self._posy

    @posy.setter
    def posy(self, value):
        """Return posy bump value."""
        self._posy = value

    @property
    def angy(self):
        """Return angy bump value."""
        return self._angy

    @angy.setter
    def angy(self, value):
        """Return angy bump value."""
        self._angy = value

    def conv_orbitcorr_kick2curr(self, kicks):
        """."""
        # 127 µs ± 1.01 µs per loop
        kickx = kicks[:self.ffwdcalc.nr_chs]
        kicky = kicks[self.ffwdcalc.nr_chs:]
        curr_chs = self._strenconv_chs.conv_strength_2_current(kickx)
        curr_cvs = self._strenconv_cvs.conv_strength_2_current(kicky)
        currs = _np.hstack((curr_chs, curr_cvs))
        return currs

    def bump_get_orbitcorr_current(self):
        """Return bump orbit correctors currents."""
        kicks = self.ffwdcalc.conv_posang2kick(
            self.posx, self.angx, self.posy, self.angy)
        currents = self.conv_orbitcorr_kick2curr(kicks)
        return currents

    def ffwd_get_orbitcorr_current(self, phase=None):
        """Return feedforward orbitcorr currents for a given ID phase."""
        # 157 µs ± 3.93 µs per loop [if phase is passed]
        if phase is None:
            phase = self.apu.phase
        currents = self.ffwdcalc.conv_phase_2_orbcorr_currents(phase)
        return currents

    def ffwd_update_orbitcorr(self, phase=None):
        """Update orbit feedforward."""
        currents_ffwd = self.ffwd_get_orbitcorr_current(phase)
        currents_bump = self.bump_get_orbitcorr_current()
        self.correctors.orbitcorr_current = currents_ffwd + currents_bump

    def ffwd_update(self, phase=None):
        """Update feedforward with bump."""
        # 305 µs ± 45.5 µs per loop
        self.ffwd_update_orbitcorr(phase)

    # --- private methods ---

    def _init_posang(self):
        _ = self  # throaway arguments (temporary)
        posx, angx = 0.0, 0.0
        posy, angy = 0.0, 0.0
        # NOTE: we could initialize posang with corrector values.
        return posx, angx, posy, angy

    def _create_strenconv(self):
        """."""
        psnames = self.correctors.orbitcorr_psnames

        maname = psnames[0].replace(':PS-', ':MA-')
        strenconv_chs = _StrengthConv(
            maname, proptype='Ref-Mon', auto_mon=True)
        maname = psnames[self.ffwdcalc.nr_chs].replace(':PS-', ':MA-')
        strenconv_cvs = _StrengthConv(
            maname, proptype='Ref-Mon', auto_mon=True)

        return strenconv_chs, strenconv_cvs



"""."""

import multiprocessing as _mp

from epics import ca as _ca


# NOTE: I have to rederive epics.CAProcess here to ensure the process will be
# launched with the spawn method.
class CAProcessSpawn(_mp.get_context('spawn').Process):
    """
    A Channel-Access aware (and safe) subclass of multiprocessing.Process
    Use CAProcess in place of multiprocessing.Process if your Process will
    be doing CA calls!
    """
    def __init__(self, **kws):
        """."""
        super().__init__(**kws)

    def run(self):
        """."""
        _ca.initial_context = None
        _ca.clear_cache()
        super().run()



"""Devices of Archive PVs."""

from copy import deepcopy as _dcopy

import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..search import BPMSearch as _BPMSearch, PSSearch as _PSSearch

from . import exceptions as _exceptions
from .pvarch import PVDataSet as _PVDataSet
from .time import Time as _Time


class BaseDevice(_PVDataSet):
    """."""

    _CONV = 0.001

    DEVICES = _get_namedtuple('Devices', ('a', ))
    PROPTIES = _get_namedtuple('Propties', ('a', ))

    Time = _Time

    def __init__(self, devname, propty='', connector=None):
        """."""
        if devname in self.DEVICES:
            self._devname = self.DEVICES._fields[devname]
        elif devname in self.DEVICES._fields:
            self._devname = devname
        else:
            ValueError('Wrong value for devname')
        if propty in self.PROPTIES:
            self._propty = propty
        elif propty in self.PROPTIES._fields:
            idx = self.PROPTIES._fields.index(propty)
            self._propty = self.PROPTIES[idx]
        else:
            ValueError('Wrong value for propty')
        self._devnames, pvnames = self._get_pvnames()
        self._times = None
        self._values = None
        super().__init__(pvnames, connector=connector)
        self._parallel_query_bin_interval = 3600

    @property
    def devnames(self):
        """Return ordered list of device names."""
        return _dcopy(self._devnames)

    @property
    def timestamps(self):
        """Return retrieved orbit generated timestamps."""
        return self._times

    @property
    def values(self):
        """Return retrieved orbit interpolated values."""
        return self._values

    def update(self, mean_sec=None, parallel=True):
        """Update state by retrieving data."""
        super().update(mean_sec=mean_sec, parallel=parallel)

        # interpolate data
        self._times, self._values = self._interpolate_data(mean_sec)

    # --- private methods ---

    def _get_pvnames(self):
        devnames = []
        pvnames = []
        return devnames, pvnames

    def _interpolate_data(self, mean_sec):
        # calc mean_sec if not passed
        nr_pvs = len(self._pvdata)
        if mean_sec is None:
            mean_sec = sum(map(
                lambda p: p.timestamp[1]-p.timestamp[0],
                self._pvdata.values()))
            mean_sec /= nr_pvs

        # times vector
        t0_, t1_ = self.timestamp_start, self.timestamp_stop
        times = _np.arange(t0_, t1_, mean_sec)

        # builds orbit matrix using interpolation
        values = _np.zeros((len(times), nr_pvs), dtype=float)
        for i, pvname in enumerate(self._pvnames):
            pvdata = self._pvdata[pvname]
            values[:, i] = _np.interp(times, pvdata.timestamp, pvdata.value)
        values *= Orbit._CONV

        # return data
        return times, values

    def __getitem__(self, index):
        """Return raw timestamp and value tuple data for a given BPM."""
        if isinstance(index, int):
            pvname = self._pvnames[index]
        elif isinstance(index, str):
            if index in self._pvnames:
                pvname = index
            elif index in self._devnames:
                index = self._devnames.index(index)
                pvname = self._pvnames[index]
            else:
                raise _exceptions.IndexError
        pvdata = self._pvdata[pvname]
        return pvdata.timestamp, pvdata.value * Orbit._CONV


class Orbit(BaseDevice):
    """."""

    _CONV = 0.001  # [nm -> um]

    DEVICES = _get_namedtuple(
        'Devices', ('TB_BPM', 'TS_BPM', 'BO_BPM', 'SI_BPM', 'SI_PBPM'))
    PROPTIES = _get_namedtuple(
        'Propties',
        ('PosX', 'PosY', 'PosQ', 'Sum'),
        ('PosX-Mon', 'PosY-Mon', 'PosQ-Mon', 'Sum-Mon'))

    def __init__(self, devname='SI_BPM', propty='PosX-Mon', connector=None):
        """."""
        super().__init__(devname, propty=propty, connector=connector)

    # --- private methods ---
    def _get_pvnames(self):
        sec, dev = self._devname.split('_')
        devnames = _dcopy(_BPMSearch.get_names({'sec': sec, 'dev': dev}))
        pvnames = tuple(dev+':'+self._propty for dev in devnames)
        return devnames, pvnames


class Correctors(BaseDevice):
    """."""

    _CONV = 1  # [urad -> urad]

    DEVICES = _get_namedtuple(
        'Devices',
        (
            'TB_CH', 'TB_CV', 'TS_CH', 'TS_CV',
            'BO_CH', 'BO_CV', 'SI_CH', 'SI_CV', 'SI_FCH', 'SI_FCV'
        )
    )
    PROPTIES = _get_namedtuple(
        'Propties',
        (
            'Curr_Mon', 'Curr_Ref', 'Curr_SP', 'Curr_RB',
            'Kick_Mon', 'Kick_Ref', 'Kick_SP', 'Kick_RB',),
        (
            'Current-Mon', 'CurrentRef-Mon', 'Current-SP', 'Current-RB',
            'Kick-Mon', 'KickRef-Mon', 'Kick-SP', 'Kick-RB'))

    def __init__(self, devname='SI_CH', propty='Current-Mon', connector=None):
        """."""
        super().__init__(devname, propty=propty, connector=connector)

    # --- private methods ---
    def _get_pvnames(self):
        sec, dev = self._devname.split('_')
        devnames = _PSSearch.get_psnames({'sec': sec, 'dev': dev})
        pvnames = tuple(dev+':'+self._propty for dev in devnames)
        return devnames, pvnames


class TrimQuads(BaseDevice):
    """."""

    _CONV = 1  # [urad -> urad]

    DEVICES = _get_namedtuple(
        'Devices', ('TB_QN', 'TS_QN', 'SI_QN', 'SI_QS'))
    PROPTIES = _get_namedtuple(
        'Propties',
        (
            'Curr_Mon', 'Curr_Ref', 'Curr_SP', 'Curr_RB',
            'KL_Mon', 'KL_Ref', 'KL_SP', 'KL_RB',),
        (
            'Current-Mon', 'CurrentRef-Mon', 'Current-SP', 'Current-RB',
            'KL-Mon', 'KLRef-Mon', 'KL-SP', 'KL-RB'))

    def __init__(self, devname='SI_QS', propty='Current-Mon', connector=None):
        """."""
        super().__init__(devname, propty=propty, connector=connector)

    # --- private methods ---
    def _get_pvnames(self):
        sec, dev = self._devname.split('_')
        if dev == 'QN':
            devnames = _PSSearch.get_psnames({'sec': sec, 'dev': 'Q[DF1-9]'})
            devnames = [dev for f in devnames if 'Fam' not in dev]
        else:
            devnames = _PSSearch.get_psnames({'sec': sec, 'dev': dev})
        pvnames = tuple(dev+':'+self._propty for dev in devnames)
        return devnames, pvnames



#!/usr/local/bin/python-sirius
"""Driver module."""

from pcaspy import Alarm as _Alarm, Severity as _Severity

from ...namesys import SiriusPVName
from ..app import App as _App
from ..pvs import ComputedPV as _ComputedPV

from .csdev import conv_dev_2_liname, Const as _Const
from .pvs import LIScalarDiffPV as _LIScalarDiffPV, \
    LIVecDiffPV as _LIVecDiffPV, \
    LIRFStatusPV as _LIRFStatusPV, \
    LIPUStatusPV as _LIPUStatusPV, \
    LIEGHVStatusPV as _LIEGHVStatusPV, \
    LIFilaPSStatusPV as _LIFilaPSStatusPV


class LIDiagApp(_App):
    """Main application responsible for updating DB."""

    def _create_computed_pvs(self, *args):
        pref = self._prefix + ('-' if self._prefix else '')

        # # RF devices
        for dev in _Const.RF_DEVICES:
            devname = SiriusPVName(dev)
            prefliname = pref + conv_dev_2_liname(dev)
            prefname = devname.substitute(prefix=self._prefix)

            # DiagAmpDiff-Mon
            pvs = [None]*2
            pvs[_LIScalarDiffPV.SP] = prefliname + ':SET_AMP'
            pvs[_LIScalarDiffPV.RB] = prefliname + ':GET_AMP'
            pvo = _ComputedPV(
                devname + ':DiagAmpDiff-Mon', _LIScalarDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagPhaseDiff-Mon
            pvs = [None]*2
            pvs[_LIScalarDiffPV.SP] = prefliname + ':SET_PHASE'
            pvs[_LIScalarDiffPV.RB] = prefliname + ':GET_PHASE'
            pvo = _ComputedPV(
                devname + ':DiagPhaseDiff-Mon', _LIScalarDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagIxQDiff-Mon
            pvs = [None]*4
            pvs[_LIVecDiffPV.I_SETT] = prefliname + ':GET_CH1_SETTING_I'
            pvs[_LIVecDiffPV.Q_SETT] = prefliname + ':GET_CH1_SETTING_Q'
            pvs[_LIVecDiffPV.I_DATA] = prefliname + ':GET_CH1_I'
            pvs[_LIVecDiffPV.Q_DATA] = prefliname + ':GET_CH1_Q'
            pvo = _ComputedPV(
                devname + ':DiagIxQDiff-Mon', _LIVecDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagStatus-Mon
            pvs = [None]*7
            pvs[_LIRFStatusPV.PV_STREAM] = prefliname + ':GET_STREAM'
            pvs[_LIRFStatusPV.PV_EXTTRG] = \
                prefliname + ':GET_EXTERNAL_TRIGGER_ENABLE'
            pvs[_LIRFStatusPV.PV_INTEGR] = prefliname + ':GET_INTEGRAL_ENABLE'
            pvs[_LIRFStatusPV.PV_FBMODE] = prefliname + ':GET_FB_MODE'
            pvs[_LIRFStatusPV.PV_AMPDIF] = prefname + ':DiagAmpDiff-Mon'
            pvs[_LIRFStatusPV.PV_PHSDIF] = prefname + ':DiagPhaseDiff-Mon'
            pvs[_LIRFStatusPV.PV_IXQDIF] = prefname + ':DiagIxQDiff-Mon'
            pvo = _ComputedPV(
                devname + ':DiagStatus-Mon', _LIRFStatusPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

        # # PU devices
        for dev in _Const.PU_DEVICES:
            devname = SiriusPVName(dev)
            prefliname = pref + conv_dev_2_liname(dev)

            # DiagVoltageDiff-Mon
            pvs = [None]*2
            pvs[_LIScalarDiffPV.SP] = prefliname + ':WRITE_V'
            pvs[_LIScalarDiffPV.RB] = prefliname + ':READV'
            pvo = _ComputedPV(
                devname + ':DiagVoltageDiff-Mon', _LIScalarDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagCurrentDiff-Mon
            pvs = [None]*2
            pvs[_LIScalarDiffPV.SP] = prefliname + ':WRITE_I'
            pvs[_LIScalarDiffPV.RB] = prefliname + ':READI'
            pvo = _ComputedPV(
                devname + ':DiagCurrentDiff-Mon', _LIScalarDiffPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

            # DiagStatus-Mon
            pvs = [None]*17
            pvs[_LIPUStatusPV.PV.RUNSTOP] = prefliname + ':Run/Stop'
            pvs[_LIPUStatusPV.PV.PREHEAT] = prefliname + ':PreHeat'
            pvs[_LIPUStatusPV.PV.CHRALLW] = prefliname + ':Charge_Allowed'
            pvs[_LIPUStatusPV.PV.TRGALLW] = prefliname + ':TrigOut_Allowed'
            pvs[_LIPUStatusPV.PV.EMRSTOP] = prefliname + ':Emer_Stop'
            pvs[_LIPUStatusPV.PV.CPS_ALL] = prefliname + ':CPS_ALL'
            pvs[_LIPUStatusPV.PV.THYHEAT] = prefliname + ':Thy_Heat'
            pvs[_LIPUStatusPV.PV.KLYHEAT] = prefliname + ':Kly_Heat'
            pvs[_LIPUStatusPV.PV.LVRDYOK] = prefliname + ':LV_Rdy_OK'
            pvs[_LIPUStatusPV.PV.HVRDYOK] = prefliname + ':HV_Rdy_OK'
            pvs[_LIPUStatusPV.PV.TRRDYOK] = prefliname + ':TRIG_Rdy_OK'
            pvs[_LIPUStatusPV.PV.SELFFLT] = prefliname + ':MOD_Self_Fault'
            pvs[_LIPUStatusPV.PV.SYS_RDY] = prefliname + ':MOD_Sys_Ready'
            pvs[_LIPUStatusPV.PV.TRGNORM] = prefliname + ':TRIG_Norm'
            pvs[_LIPUStatusPV.PV.PLSCURR] = prefliname + ':Pulse_Current'
            pvs[_LIPUStatusPV.PV.VLTDIFF] = prefliname + ':DiagVoltageDiff-Mon'
            pvs[_LIPUStatusPV.PV.CRRDIFF] = prefliname + ':DiagCurrentDiff-Mon'
            pvo = _ComputedPV(
                devname + ':DiagStatus-Mon', _LIPUStatusPV(),
                self._queue, pvs, monitor=False)
            self.pvs.append(pvo)

        # # Egun devices
        # HVPS
        devname = SiriusPVName(_Const.HVPS)
        prefliname = pref + conv_dev_2_liname(devname)
        prefname = devname.substitute(prefix=self._prefix)
        # DiagVoltDiff-Mon
        pvs = [None]*2
        pvs[_LIScalarDiffPV.SP] = prefliname + ':voltoutsoft'
        pvs[_LIScalarDiffPV.RB] = prefliname + ':voltinsoft'
        pvo = _ComputedPV(
            devname + ':DiagVoltDiff-Mon', _LIScalarDiffPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)
        # DiagStatus-Mon
        pvs = [None]*3
        pvs[_LIEGHVStatusPV.PV_SWITCH] = prefliname + ':swstatus'
        pvs[_LIEGHVStatusPV.PV_ENABLE] = prefliname + ':enstatus'
        pvs[_LIEGHVStatusPV.PV_VLTDIF] = prefname + ':DiagVoltDiff-Mon'
        pvo = _ComputedPV(
            devname + ':DiagStatus-Mon', _LIEGHVStatusPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        # FilaPS
        devname = SiriusPVName(_Const.FILA)
        prefliname = pref + conv_dev_2_liname(devname)
        prefname = devname.substitute(prefix=self._prefix)
        # DiagCurrentDiff-Mon
        pvs = [None]*2
        pvs[_LIScalarDiffPV.SP] = prefliname + ':currentoutsoft'
        pvs[_LIScalarDiffPV.RB] = prefliname + ':currentinsoft'
        pvo = _ComputedPV(
            devname + ':DiagCurrentDiff-Mon', _LIScalarDiffPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)
        # DiagStatus-Mon
        pvs = [None]*2
        pvs[_LIFilaPSStatusPV.PV_SWITCH] = prefliname + ':swstatus'
        pvs[_LIFilaPSStatusPV.PV_CURDIF] = prefname + ':DiagCurrentDiff-Mon'
        pvo = _ComputedPV(
            devname + ':DiagStatus-Mon', _LIFilaPSStatusPV(),
            self._queue, pvs, monitor=False)
        self.pvs.append(pvo)

        self._pvs_connected = {pv: False for pv in self.pvs}

    def _update_pvs(self):
        for pvo in self.pvs:
            if not pvo.connected:
                if self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.TIMEOUT_ALARM,
                        severity=_Severity.INVALID_ALARM,
                        field='status')
                self._pvs_connected[pvo] = False
                if 'DiagStatus' in pvo.pvname:
                    self.run_callbacks(pvo.pvname, value=pvo.value)
            else:
                if not self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.NO_ALARM,
                        severity=_Severity.NO_ALARM, field='status')
                self._pvs_connected[pvo] = True
                self.run_callbacks(pvo.pvname, value=pvo.value)
